import logging
import tempfile
import unittest, os
# ======================================================================================================================
from multiprocessing import Process

import pyrosetta

# ======================================================================================================================
from rdkit import Chem
from rdkit.Chem import AllChem, rdqueries
from rdkit.Geometry import Point3D

from fragmenstein import Monster, Victor, Igor, mpro_data, Walton
from fragmenstein.mpro import MProVictor
from typing import *
import numpy as np

# ======================================================================================================================

class Internals(unittest.TestCase):
    def test_triangle(self):
        """
        Test triangle prevention.
        """
        probanda = Chem.MolFromSmiles('CCC')
        AllChem.EmbedMolecule(probanda)
        monster = Monster([probanda])
        zeroth = probanda.GetAtomWithIdx(0)
        second = probanda.GetAtomWithIdx(2)
        # 0 - 2 would make a triangle
        self.assertTrue(
            monster._is_would_be_triangle(zeroth, second))  # connecting zeroth and second would make a triangle
        self.assertEqual(monster._get_triangle(zeroth, second), 1)  # connecting 0+2, would make 1 the vertex.

    def make_mol(self, smiles: str) -> Chem.Mol:
        mol = Chem.MolFromSmiles(smiles)
        dummies = mol.GetAtomsMatchingQuery(rdqueries.AtomNumEqualsQueryAtom(0))
        for dummy in dummies:
            dummy.SetAtomicNum(6)  # carbon is 6 not 12!
        AllChem.EmbedMolecule(mol)
        for dummy in dummies:
            dummy.SetAtomicNum(0)
        return mol

    def make_pair_by_split(self, conjoined: Chem.Mol, atom_idx: int) -> Tuple[Chem.Mol, Chem.Mol]:
        # make overlapping mols by getting a single molecule, and split it
        # this gives more control over Chem.rdMolAlign.AlignMol as this may overlap other atoms.
        # negative weights does not work...
        # fore
        bond = conjoined.GetBondBetweenAtoms(atom_idx, atom_idx + 1)
        fragged = Chem.FragmentOnBonds(conjoined, [bond.GetIdx()], addDummies=False)
        fore = Chem.GetMolFrags(fragged, asMols=True)[0]
        bond = conjoined.GetBondBetweenAtoms(atom_idx - 1, atom_idx)
        fragged = Chem.FragmentOnBonds(conjoined, [bond.GetIdx()], addDummies=False)
        aft = Chem.GetMolFrags(fragged, asMols=True)[1]
        return fore, aft

    def test_merge_on_same_dummy(self):
        conjoined = self.make_mol('O=C(O)CSCC#N')
        acetyl, nitrile = self.make_pair_by_split(conjoined, 4)
        # merge
        monster = Monster([acetyl, nitrile])
        merger = monster.simply_merge_hits()
        dummies = merger.GetAtomsMatchingQuery(Chem.rdqueries.AtomNumEqualsQueryAtom(0))
        for dummy in dummies:
            self.assertEqual(len(dummy.GetNeighbors()), 1)
        self.assertEqual(len(Chem.GetMolFrags(merger)), 1)

    def translate(self, mol, x=0, y=0, z=0):
        """
        Translates the molecule in place

        :param mol:
        :param x:
        :param y:
        :param z:
        :return:
        """
        translation = np.array([[1, 0, 0, x],
                                [0, 1, 0, y],
                                [0, 0, 1, z],
                                [0, 0, 0, 1]], dtype=np.double)
        AllChem.TransformConformer(mol.GetConformer(0), translation)

    def test_join(self):
        wanted = 'c1ccc(Oc2ccccc2)cc1'
        benzene = Chem.MolFromSmiles('c1ccccc1')
        AllChem.EmbedMolecule(benzene)
        moved = Chem.Mol(benzene)
        self.translate(moved, x=5)
        mol = Monster([benzene, moved]).combine().positioned_mol
        found = Chem.MolToSmiles(Chem.RemoveHs(mol))
        self.assertEqual(wanted, found, 'The joining differs')

    def test_distance(self):
        methane = Chem.MolFromSmiles('C')
        AllChem.EmbedMolecule(methane)
        ammonia = Chem.MolFromSmiles('N')
        AllChem.EmbedMolecule(ammonia)
        self.translate(ammonia, x=3)
        monster = Monster([methane, ammonia])
        monster.combine(keep_all=False, joining_cutoff=2)
        self.assertEqual(1, len(monster.unmatched), 'discard error')
        monster.combine(keep_all=True, joining_cutoff=5)
        self.assertEqual(0, len(monster.unmatched), 'discard error')
        try:
            monster.combine(keep_all=True, joining_cutoff=2)
            self.fail('should have raised a connection error')
        except ConnectionError as error:
            pass

    def test_neigh_bonding(self):
        """
        This was to address the bug from 5/1/23.

        Issue was in ``_nan_fill_submatrix``, which was not blanking due to
        ``isinstance(np.int64(42), int) == False``
        """
        # make mols
        methane: Chem.Mol = Chem.MolFromSmiles('C')
        methane.SetProp('_Name', 'methane')
        AllChem.EmbedMolecule(methane, coordMap={0: Point3D(0, 0, 0)})
        ammonia = Chem.MolFromSmiles('N')
        ammonia.SetProp('_Name', 'ammonia')
        # merge
        monster = Monster([])
        AllChem.EmbedMolecule(ammonia, coordMap={0: Point3D(3, 0, 0)})
        self.translate(ammonia, x=3)
        # monster.join_neighboring_mols(methane, ammonia)
        combo, candidates = monster._find_all_closest(methane, ammonia)  # _find_all_closest is in communal
        self.assertEqual(combo.GetNumAtoms(), 2)
        self.assertNotEqual(candidates[0][0], candidates[0][1])



if __name__ == '__main__':
    unittest.main()



import logging
import pebble
import operator
from typing import (Any, Callable, Union, Iterator, Sequence, List)
from collections import Counter
import pandas as pd
from rdkit import Chem
from ..monster import Monster
from ..victor import Victor


def binarize(mol:Chem.Mol, ignore_errors:bool=True) -> bytes:
    # this is convoluted as None is a common error outcome with RDKit
    # so making it less cryptic
    if isinstance(mol, bytes):
        return mol   # messed up. it was bytes.
    elif not isinstance(mol, Chem.Mol): # likely None
        if ignore_errors:
            mol = Chem.Mol()    # noqa
        elif mol is None:
            raise TypeError('A Chem.Mol was expected by got None. '+
                            'RDKit returns None if it fails an operation.')
        else:
            raise TypeError('the molecule provide is not a Chem.Mol, '+
                            f'but a {type(mol).__name__}')
    exception = Exception if ignore_errors else ()
    try:
        return mol.ToBinary(propertyFlags=0b00010111)    # noqa it's an instance.
    except exception as error:
        return Chem.Mol().ToBinary(propertyFlags=0b00010111)  # noqa


def unbinarize(bin: bytes, ignore_errors:bool=True) -> Union[Chem.Mol, None]:
    exception = Exception if ignore_errors else ()
    if isinstance(bin, Chem.Mol):  # lol
        return bin
    if isinstance(bin, float):  # nan
        return None
    if bin is None:
        return None
    try:
        return Chem.Mol(bin)
    except exception:
        return None

class LabBench:

    # the ``outcome`` column in the pandas dataframe can have these values in order of niceness:
    category_labels = ['crashed', 'too distant', 'timeout', 'unstable', 'equally sized', 'deviant', 'acceptable']

    def __init__(self, pdbblock: str, covalent_resi: Union[int, str, None]):
        self.pdbblock = pdbblock
        self.covalent_resi = covalent_resi
        self.init_options = '-ex1 -ex2 -no_optH false -mute all -ignore_unrecognized_res true -load_PDB_components false'
        self.raw_results = []
        if not len(Victor.journal.handlers):
            Victor.enable_stdout(logging.CRITICAL)

    # keep it just in case: (not called within class as that would require `self.__class__.binarize`)
    binarize: Callable[[Chem.Mol, bool], bytes] = staticmethod(binarize)
    unbinarize: Callable[[bytes, bool], Union[Chem.Mol, None]] = staticmethod(unbinarize)

    def get_completed(self, futures: pebble.ProcessMapFuture) -> pd.DataFrame:
        """
        A pebble.ProcessMapFuture cannot be stored as an instance attribute as it's not picklable
        as it handles the lock. As a result it must be passed.

        Fills ``self.raw_results`` with the results of the futures, before returning a dataframe.
        """
        result_iter: Iterator[int] = futures.result()
        self.raw_results = []
        while True:
            try:
                result = next(result_iter)
                self.raw_results.append(result)
            except TimeoutError as error:
                print("Function took longer than %d seconds" % error.args[1])
                self.raw_results.append({'error': 'TimeoutError', 'name': ''})
            except StopIteration as error:
                # it would be nice having a mock entry with the expected values...
                # self.raw_results.append({})
                break
            except KeyboardInterrupt as error:
                print('Keyboard!')
                self.raw_results.append({'error': 'KeyboardInterrupt', 'name': ''})
                break
            except Exception as error:
                print(f'{error.__class__.__name__}: {error}')
                self.raw_results.append({'error': error.__class__.__name__, 'name': ''})
        # list of dict to dataframe
        df = pd.DataFrame(self.raw_results)
        df['LE'] = df.apply(
            lambda row: row['∆∆G'] / (row.N_constrained_atoms + row.N_unconstrained_atoms),
            axis=1)
        nan_to_list = lambda value: value if isinstance(value, list) else []
        df['disregarded'] = df.disregarded.apply(nan_to_list)  # str
        df['regarded'] = df.regarded.apply(nan_to_list)  # str
        df['unminimized_mol'] = df.unmin_binary.apply(unbinarize)
        df['minimized_mol'] = df.min_binary.apply(unbinarize)
        df['hit_mols'] = df.hit_binaries.apply(lambda l: [unbinarize(b) for b in l] if isinstance(l, Sequence) else [])
        df['outcome'] = df.apply(self.categorize, axis=1)
        df['percent_hybrid'] = df.unminimized_mol.apply(self.percent_hybrid)
        return df

    def categorize(self, row: pd.Series) -> str:
        """
        Given a row categorise the 'outcome' field.
        Called by ``get_completed``.

        Subclass Laboratory to change what the outcome field ends up being.
        But do remember to update the ``category_labels`` attribute.
        """
        # see category_labels for list of values.
        is_filled: Callable[[Any], int] = lambda value: len(value) != 0 if hasattr(value, '__len__') else False
        if is_filled(row.disregarded) or 'ConnectionError' in row.error:
            # either is the same thing, but the Victor error catching makes them different
            return 'too distant'
        elif 'TimeoutError' in row.error:
            return 'timeout'
        elif is_filled(row.error):
            return 'crashed'
        elif max(map(operator.methodcaller('GetNumHeavyAtoms'), row.hit_mols))  \
                >= row.unminimized_mol.GetNumHeavyAtoms():
            return 'equally sized'
        elif row.comRMSD > 1:
            return 'too moved'
        elif row['∆∆G'] >= 0:
            return 'too contorted'
        else:
            return 'acceptable'

    def percent_hybrid(self, mol: Chem.Mol) -> float:
        """
        Given the origins how much of the molecule is not accounted for by the primary hit?

        :param mol:
        :return:
        """
        if not isinstance(mol, Chem.Mol):
            return float('nan')
        origins: List[List[str]] = Monster.origin_from_mol(None, mol)
        c = Counter([o[0].split('.')[0] if o else None for o in origins]).most_common()
        return round((mol.GetNumHeavyAtoms() - c[0][1]) / mol.GetNumHeavyAtoms() * 100, 1)

    def __call__(self,
                 iterator: Iterator,
                 fun: Callable,
                 n_cores: int = 4,
                 timeout: int = 240,
                 asynchronous: bool = False
                 ):
        """Combine/permute the molecules ``mols``
        on ``n_cores`` subprocesses.
        killing any that live longer than ``timeout`` seconds.
        The method returns an iterator of promises ``pebble.ProcessMapFuture`` if ``asynchronous`` is True,
        or the results as a pandas DataFrame. To convert the promises to a dataframe use ``get_completed``."""
        with pebble.ProcessPool(max_workers=n_cores, max_tasks=n_cores) as pool:
            futures: pebble.ProcessMapFuture = pool.map(fun, iterator, timeout=timeout)
        if asynchronous:
            return futures
        else:
            return self.get_completed(futures)



from ._no_blending import _MonsterNone
from typing import Dict, List, Tuple, Optional, Unpack, Set  # noqa: F401
from ..positional_mapping import GPM
import itertools
from ..mcs_mapping import IndexMap, ExtendedFMCSMode
from copy import deepcopy
from rdkit import Chem
from ..unmerge_mapper import Unmerge

class _MonsterExpand(_MonsterNone):
    """
    A variant of no_blend mode, with a focus on expansion
    """
    def by_expansion(self, primary_name:Optional[str]=None, min_mode_index:int=0) -> Chem.Mol:
        """
        Get the maps. Find the map with the most atoms covered.
        Use that map as the base map for the other maps.
        """
        # -------------- Get the primary hit ----------------------------
        primary_maps: List[Dict[int, int]]
        # primary_name as None chooses one, else the primary name provided is used:
        primary_name, primary_maps = self._get_primary_maps(primary_name)
        # -------------- Get the secondary hits --------------------------------
        # positional_overlap is used by ``_expand_hit_atom_map_by_overlap``
        # which is called by ``_get_unmerge_expansions``
        positional_overlaps: Dict[Tuple[str, str], Dict[int, int]] = self._compute_overlaps()
        unmergers: List[Unmerge] = self._get_unmerge_expansions(primary_name,
                                                                primary_maps,
                                                                positional_overlaps,
                                                                min_mode_index)
        # -------------- Sort the unmergers --------------------------------
        self.positioned_mol, self.mol_options = self._place_unmerger_expansions(unmergers)
        return self.positioned_mol

    def _get_primary_maps(self, primary_name:Optional[str]=None) -> Tuple[str, List[Dict[int, int]]]:
        """
        The primary hit is the hit will most in common with the placed molecule.

        :param primary_name:
        :return:
        """
        if primary_name is None:
            # the list is [{hit_atom_idx: template_atom_idx}, ...]
            maps: Dict[str, List[Dict[int, int]]] = self._compute_maps(broad=True)
            # get the largest maps (not the number of maps which would be `len(l)`)
            get_size = lambda l: len(l[0]) if len(l) else 0  # noqa: E731 Guido doesn't like lambda, but I do
            max_size = max(map(get_size, maps.values()))
            # sorted_maps: Dict[str, List[Dict[int, int]]] = dict(sorted(maps.items(),
            #                                                            key=lambda x: get_size(x[1]),
            #                                                            reverse=True))
            biggest_maps = {k: v for k, v in maps.items() if get_size(v) == max_size}
            # choose the first map
            primary_name = list(biggest_maps.items())[0][0]
            primary_maps: List[Dict[int, int]] = biggest_maps[primary_name]
        else:
            primary: Chem.Mol = self.get_hit_by_name(primary_name)
            primary_maps: List[Dict[int, int]] = self._compute_hit_maps(primary, broad=True)
        self.journal.debug(f"Primary hit: {primary_name} with {len(primary_maps)} Primary maps: {primary_maps}")
        return primary_name, primary_maps

    def _get_unmerge_expansions(self, primary_name, primary_maps,
                                positional_overlaps,
                                min_mode_index) -> List[Unmerge]:
        """
        Calls _perform_unmerge which calls Unmerge
        """
        unmergers = []
        # the no_blend mode does the unmerged based on a dict of optional maps,
        # i.e. the maps do not affect each other. Here it is important that they do.
        # hence each primary map is converted into a set of unmerge maps and the best wins.
        # if there is only one hit, then the primary map is the only unmerge map...
        if len(self.hits) == 1:
            return [self._perform_unmerge(maps={primary_name: primary_maps},
                                          n_poisonous=3,
                                          primary_name=primary_name
                                          )]
        # case: multiple hits
        for primary_map in primary_maps:  #: Dict[int, int]
            # iterate over the hit map and expand to all overlapping atoms
            self.journal.debug(f'primary_map: {primary_map}')
            exp_map: Dict[str, Dict[int, int]] = self._expand_hit_atom_map_by_overlap(primary_name,
                                                                                      primary_map,
                                                                                      positional_overlaps,
                                                                                      self.custom_map)
            self.journal.debug(f'initial expanded map (primary + overlaps): {exp_map}')
            exp_maps = {primary_name: [primary_map]}  # only one primary map!
            accounted_for: Set[int] = {i for i in primary_map.values() if i >= 0}
            # get the maps that are not the primary map
            for other in self.hits:
                other_name: str = other.GetProp('_Name')
                if other_name == primary_name:
                    continue
                mappings:List[Dict[int, int]]
                mode: ExtendedFMCSMode
                mappings, mode = self.get_mcs_mappings(other, self.initial_mol, min_mode_index, exp_map)
                # drop any that are redundant with the primary hit
                mappings = [d for d in mappings if len(set(d.values()) - accounted_for) > 0]
                exp_maps[other_name] = mappings
                self.journal.debug(f'candiate expanded maps: {exp_maps} following: {other_name}')
            # {h: f for h, f in .items() if h >= 0 and f >= 0}
            unmergers.append(self._perform_unmerge(maps=exp_maps,
                                                   n_poisonous=3,
                                                   primary_name=primary_name))
        return unmergers

    def _place_unmerger_expansions(self, unmergers: List[Unmerge]) -> Tuple[Chem.Mol, List[Chem.Mol]]:
        scores: List[int] = []
        mol_options = []
        best_mol = None
        for unmerger in unmergers:
            n_off_atoms: int = unmerger.offness(unmerger.combined, unmerger.combined_map)
            scores.append(len(unmerger.combined_map) - 3 * n_off_atoms)
        max_score = max(scores)
        # if they came out equal keep both...
        for score, unmerger in zip(scores, unmergers):
            if score != max_score:
                continue
            positioned_mol, inner_options = self._place_unmerger(unmerger)
            mol_options.extend(inner_options)
            if best_mol:  # it might get sorted again... so it is not important
                mol_options.insert(0, positioned_mol)
            else:
                best_mol = positioned_mol
        return best_mol, mol_options

    def _compute_overlaps(self) -> Dict[Tuple[str, str], Dict[int, int]]:
        positional_overlaps: Dict[Tuple[str, str], Dict[int, int]] = {}
        for mol1, mol2 in itertools.combinations(self.hits, 2):
            mol1_name:str = mol1.GetProp('_Name')
            mol2_name:str = mol2.GetProp('_Name')
            gpm = GPM.get_positional_mapping(mol1, mol2)
            positional_overlaps[(mol1_name, mol2_name)] = gpm
            positional_overlaps[(mol2_name, mol1_name)] = gpm
        return positional_overlaps


    def _include_missing_hits(self, custom_map: Dict[str, Dict[int, int]]) -> None:
        for hit in self.hits:
            name = hit.GetProp('_Name')
            if name not in custom_map:
                custom_map[name] = {}

    def _expand_hit_atom_map_by_overlap(self,
                                     hit_name: str,
                                     hit_atom_map: Dict[int, int],
                                     positional_overlaps: Dict[Tuple[str, str], Dict[int, int]],
                                     custom_map: Dict[str, Dict[int, int]]) -> Dict[str, Dict[int, int]]:
        """
        Expanded the custom_map by adding all atoms that are covered by the hit_atom_map.

        :param hit_name:
        :param hit_atom_map:
        :param positional_overlaps:
        :param custom_map:
        :param mode:
        :return: custom_map
        """
        expanded: Dict[str, Dict[int, int]] = deepcopy(custom_map)
        expanded[hit_name] = hit_atom_map
        self._include_missing_hits(expanded)
        for hit_atom_idx, template_atom_idx in hit_atom_map.items():
            for other in self.hits:
                other_name:str = other.GetProp('_Name')
                if other_name == hit_name:
                    continue
                # ------------- deal with atoms that overlap --------------------------
                overlaps: Dict[int, int] = positional_overlaps[(hit_name, other_name)]
                if hit_atom_idx in overlaps:
                    # ignore the special overrides
                    if template_atom_idx < 0:
                        continue
                    if hit_atom_idx < 0:
                        continue
                    if overlaps[hit_atom_idx] in expanded[other_name]:
                        continue
                    expanded[other_name][overlaps[hit_atom_idx]] = template_atom_idx
                # ------------- deal with atoms that do not overlap ------------------
                elif template_atom_idx in expanded[other_name].values():
                    pass  # there is a mapping already ?!
                else:  # damn the template_atom_idx
                    expanded[other_name][-2-hit_atom_idx] = template_atom_idx
        return expanded




from rdkit import Chem
from rdkit.Chem import rdFMCS
# TypedDict & Unpack fixed in .legacy:
from typing import Dict, List, Tuple, Optional, TypeVar, Sequence, TypedDict, Unpack  # noqa
from functools import singledispatchmethod  # monkeypatched by .legacy (for Py3.7)
import itertools

from .types import IndexMap, BasicFMCSMode, ExtendedFMCSMode


class VanillaCompareAtoms(rdFMCS.MCSAtomCompare):
    """
    Give than the following does not work, one has to do it in full:

    .. code-block:: python
        :caption: This will raise an error:
        super().__call__(parameters, mol1, atom_idx1, mol2, atom_idx2)

    This class replicates the vanilla functionality
    """

    def __init__(self, comparison: rdFMCS.AtomCompare = rdFMCS.AtomCompare.CompareAnyHeavyAtom):
        """
        Whereas the atomCompare is an enum, this is a callable class.
        But in parameters there is no compareElement booleans etc. only Isotope...
        In https://github.com/rdkit/rdkit/blob/master/Code/GraphMol/FMCS/Wrap/testFMCS.py
        it is clear one needs to make one's own.
        """
        super().__init__()  # noqa the p_obect (python object aka ctypes.py_object) is not used
        self.comparison = comparison

    def __call__(self,  # noqa signature matches... it is just Boost being Boost
                 parameters: rdFMCS.MCSAtomCompareParameters,
                 mol1: Chem.Mol,
                 atom_idx1: int,
                 mol2: Chem.Mol,
                 atom_idx2: int) -> bool:
        a1: Chem.Atom = mol1.GetAtomWithIdx(atom_idx1)
        a2: Chem.Atom = mol2.GetAtomWithIdx(atom_idx2)
        # ------- isotope ------------------------
        if parameters.MatchIsotope and a1.GetIsotope() != a2.GetIsotope():  # noqa
            return False
        elif self.comparison == rdFMCS.AtomCompare.CompareIsotopes and a1.GetIsotope() != a2.GetIsotope():  # noqa
            return False
        elif self.comparison == rdFMCS.AtomCompare.CompareElements and a1.GetAtomicNum() != a2.GetAtomicNum():  # noqa
            return False
        elif self.comparison == rdFMCS.AtomCompare.CompareAnyHeavyAtom \
                and (a1.GetAtomicNum() == 1 or a2.GetAtomicNum() == 1):  # noqa
            return False
        elif self.comparison == rdFMCS.AtomCompare.CompareAny:
            pass
        # ------- valence ------------------------
        if parameters.MatchValences and a1.GetTotalValence() != a2.GetTotalValence():  # noqa
            return False
        # ------- chiral ------------------------
        if parameters.MatchChiralTag and not self.CheckAtomChirality(parameters, mol1, atom_idx1, mol2, atom_idx2):
            return False
        # ------- formal ------------------------
        if parameters.MatchFormalCharge and not self.CheckAtomCharge(parameters, mol1, atom_idx1, mol2, atom_idx2):
            return False
        # ------- ring ------------------------
        if parameters.RingMatchesRingOnly and not self.CheckAtomRingMatch(parameters, mol1, atom_idx1, mol2, atom_idx2):
            return False
        # ------- complete ------------------------
        if parameters.CompleteRingsOnly:
            # don't know its intended to be used
            pass
        # ------- distance ------------------------
        if parameters.MaxDistance:  # integer...
            # todo fill!
            pass
        return True


class SpecialCompareAtoms(VanillaCompareAtoms):
    """
    This works like the ``_get_atom_maps`` did prior to Fragmentein version 0.9.
    The mapping as discussed in GitHub issue #23 is in the format

        mapping = { 'hit1': {1:1,2:5} 'hit2': {3:3,4:4,4:6}}

    The hit index is first, followup index is the second.
    The index `-1` for a followup index is the same as not providing the hit index,
    it is described here solely for clarity not for use.

        mapping = { 'hit1': {1:1,2:5, 3:-1} 'hit2': {3:3,4:4,4:6}}

    The index `-2` for a followup index will result in the hit atom index
    not matching any followup index.

        mapping = { 'hit1': {1:1,2:5, 3:-2} 'hit2': {3:3,4:4,4:6}}

    If ``exclusive_mapping`` argument of __init__ is True,
    then if a followup index is present in one hit, but not in a second hit,
    then no atom of the second hit will match that followup atom.
    A negative index for a hit atom index means that no atom in that hit will match the
    corresponding followup index.

        mapping = { 'hit1': {1:1,2:5,-1:3, -2: 7} 'hit2': {3:3,4:4,4:6}}

    However, a positive integer on a different hit overrides it, therefore,
    in the above followup atom 3 cannot be matched to any atom in hit1, but will match
    atom 3 in hit2. Followup atom 7 will not match either.

    .. code-block:: python

        SpecialCompareAtoms(custom_map=mapping, exclusive_mapping=True)
    """

    def __init__(self,
                 comparison: rdFMCS.AtomCompare = rdFMCS.AtomCompare.CompareAnyHeavyAtom,
                 custom_map: Optional[Dict[str, Dict[int, int]]] = None, exclusive_mapping: bool = True):
        super().__init__(comparison=comparison)  # what is p_object?
        self.custom_map = self._parse_custom_map(custom_map)  # custom as in user-defined
        self.banned = self._get_strict_banned() if exclusive_mapping else self._get_lax_banned()

    def _parse_custom_map(self, custom_map):
        """
        Make sure its Dict[str, Dict[int, int]]
        """
        if custom_map is None:
            custom_map = {}
        assert isinstance(custom_map, dict), 'User defined map has to be mol name to Dict[int, int]'
        for name, hit_map in custom_map.items():
            custom_map[name] = dict(hit_map)
        return custom_map

    def _get_strict_banned(self):
        """
        A list of followup indices that cannot be unmapped
        called if exclusive_mapping is True
        """
        return [foll_idx for mapping in self.custom_map.values() for foll_idx in mapping.values()]

    def _get_lax_flipped_map(self):
        """
        A list of followup indices that cannot be mapped as per negative hit index
        called if exclusive_mapping is False
        """
        return [foll_idx for mapping in self.custom_map.values() for hit_idx, foll_idx in mapping.items()
                if hit_idx < 0]

    def get_custom(self, hit_mol: Chem.Mol, hit_atom_idx: int) -> int:
        """
        What idx of followup corresponds to the ``hit_atom_idx`` index of ``hit_mol``?
        If nothing, -1 is returned
        """
        name: str = hit_mol.GetProp('_Name')
        if name not in self.custom_map:
            # technically impossible, bar for user altered map.
            # should this corner case be handled everywhere?
            return -1
        return self.custom_map[name].get(hit_atom_idx, -1)

    def __call__(self,
                 parameters: rdFMCS.MCSAtomCompareParameters,
                 hit: Chem.Mol,
                 hit_atom_idx: int,
                 followup: Chem.Mol,
                 followup_atom_idx: int) -> bool:
        hit_atom = hit.GetAtomWithIdx(hit_atom_idx)
        followup_atom = followup.GetAtomWithIdx(followup_atom_idx)
        symbols = {hit_atom.GetSymbol(), followup_atom.GetSymbol()}
        # ------- Custom -----------------------
        # get the user defined map target -- see docstring for bla bla
        custom: int = self.get_custom(hit, hit_atom_idx)
        if custom == followup_atom_idx:
            # it is the custom map!
            return True
        elif custom != -1:
            # a different index was given
            # or a non -1 negative number
            return False
        elif followup_atom_idx in self.banned:
            # banned is fully filled if exclusive_mapping was true during init
            # otherwise its user provided negatives
            return False
        else:
            # followup index not assigned
            pass
        # ------- Dummy ------------------------
        # dummy atom cannot match non-dummy atom:
        if '*' in symbols and len(symbols) > 1:
            return False
        # ------- protons ------------------------
        # proton cannot match non-proton:
        if 'H' in symbols and len(symbols) > 1:
            return False
        # ------- vanilla ------------------------
        return super().__call__(parameters, hit, hit_atom_idx, followup, followup_atom_idx)

    @singledispatchmethod
    def get_valid_matches(self,
                          parameters: rdFMCS.MCSAtomCompareParameters,
                          common: Chem.Mol,
                          hit: Chem.Mol,
                          followup: Chem.Mol) -> List[IndexMap]:
        """
        Returns a list of possible matches, each being a lists of tuples of hit to follow indices,
        that obey the criteria of the atomic comparison.

        This however does not check if all the atoms in custom are present.
        For that, ``Monster._validate_vs_custom`` is called in the method ``Monster.get_mcs_mappings``,
        after calling ``Monster._get_atom_maps``, which calls this method.
        (``Monster.get_mcs_mappings`` tries different matching schema,
        while ``Monster._get_atom_maps`` is for one single scheme).
        The primary reason why this is so, is that there are two tiers of requirements:

        1. The custom map must be present vs
        2. The custom map may be present, but has to be obeyed.

        parameters can be rdFMCS.MCSAtomCompareParameters or rdFMCS.MCSParameters
        """

        matches = []
        # find the common matches and make sure they match each other:
        # GetSubstructMatches is independent of AtomCompare.
        for hit_match, followup_match in itertools.product(hit.GetSubstructMatches(common, uniquify=False),
                                                           followup.GetSubstructMatches(common, uniquify=False)):
            # re `map(int, hit_match)` I do not know under what condition is it not an int...
            # but it was so in previous iteration
            for h, f in zip(map(int, hit_match), map(int, followup_match)):
                if not self(parameters=parameters,
                            hit=hit,
                            followup=followup,
                            hit_atom_idx=h,
                            followup_atom_idx=f):
                    break  # one index does not match. None should match.
            else:
                matches.append(list(zip(map(int, hit_match), map(int, followup_match))))
        # remove duplicates
        matches = list(set([tuple(sorted(m, key=lambda i: i[0])) for m in matches]))
        return matches

    @get_valid_matches.register
    def _(self,
          parameters: rdFMCS.MCSParameters,
          common: Chem.Mol,
          hit: Chem.Mol,
          followup: Chem.Mol) -> List[IndexMap]:
        return self.get_valid_matches(parameters.AtomCompareParameters, common, hit, followup)



import logging
import tempfile
import unittest, os
# ======================================================================================================================
from multiprocessing import Process

import pyrosetta

# ======================================================================================================================
from rdkit import Chem
from rdkit.Chem import AllChem

from fragmenstein import Monster, Victor, Igor, mpro_data, Walton
from fragmenstein.mpro import MProVictor
from typing import *
import numpy as np

# ======================================================================================================================

class VictorCombineTests(unittest.TestCase):

    def setUp(self):
        Igor.init_pyrosetta()

    def test_noncovalent(self):
        MProVictor.quick_reanimation = False
        victor = MProVictor.from_hit_codes(hit_codes=['x0305', 'x1249'])
        victor.combine()
        self.assertLess(victor.mrmsd.mrmsd, 1, f'RMSD great that one ({victor.mrmsd.mrmsd})')
        self.assertLess(victor.ddG, -1, f'ddG {victor.ddG}')

    def test_covalent(self):
        MProVictor.quick_reanimation = False
        victor = MProVictor.from_hit_codes(hit_codes=['x0692', 'x0305', 'x1249'])
        victor.combine()
        self.assertLess(victor.mrmsd.mrmsd, 1.2, f'RMSD great that one ({victor.mrmsd.mrmsd})')
        self.assertLess(victor.ddG, -1, f'ddG {victor.ddG}')



if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np
# ======================================================================================================================
from rdkit import Chem
from rdkit.Chem import AllChem

from fragmenstein import Monster, Victor, Igor
from fragmenstein.demo import TestSet, MPro


# ======================================================================================================================

# ======================================================================================================================

class MonsterPlaceTests(unittest.TestCase):

    def test_sample_new_conformation(self):
        smiles = "C1C2C(C=C(C=2)C(C2C=CC=C2)CNOC)C=CC=1"
        mol = Chem.MolFromSmiles(smiles)
        AllChem.EmbedMolecule(mol, randomSeed=131)
        frag_mol = Chem.FragmentOnBonds(mol, [5, 11], addDummies=False)
        hits = Chem.GetMolFrags(frag_mol, asMols=True)[:2]
        # from matplotlib import pyplot as plt
        # from rdkit.Chem import Draw
        # plt.imshow(Draw.MolsToGridImage(hits)); plt.show()
        ori_monster = Monster(hits=hits, random_seed=131)
        ori_monster.place_smiles(smiles)
        # sample_new_conformation
        seeds = [121, 23421, 1]
        for se1 in seeds:
            mol = ori_monster.sample_new_conformation(random_seed=se1)
            coords1 = mol.GetConformer().GetPositions()
            for se2 in seeds:
                mol = ori_monster.sample_new_conformation(random_seed=se2)
                coords2 = mol.GetConformer().GetPositions()
                if se1 == se2:
                    self.assertAlmostEqual(np.sum(np.abs(coords1 - coords2)), 0)
                else:
                    self.assertTrue(np.sum(np.abs(coords1 - coords2)) > 3)

    def test_random_seed(self):
        smiles = "C1C2C(C=C(C=2)C(C2C=CC=C2)CNOC)C=CC=1"
        mol = Chem.MolFromSmiles(smiles)
        AllChem.EmbedMolecule(mol, randomSeed=131)
        frag_mol = Chem.FragmentOnBonds(mol, [5, 11], addDummies=False)
        hits = Chem.GetMolFrags(frag_mol, asMols=True)[:2]
        # from matplotlib import pyplot as plt
        # from rdkit.Chem import Draw
        # plt.imshow(Draw.MolsToGridImage(hits)); plt.show()
        seeds = [121, 23421, 1]
        for se1 in seeds:
            mol = Monster(hits=hits, random_seed=se1).place_smiles(smiles).positioned_mol
            coords1 = mol.GetConformer().GetPositions()
            for se2 in seeds:
                mol = Monster(hits=hits, random_seed=se2).place_smiles(smiles).positioned_mol
                coords2 = mol.GetConformer().GetPositions()
                if se1 == se2:
                    self.assertAlmostEqual(np.sum(np.abs(coords1 - coords2)), 0)
                else:
                    self.assertTrue(np.sum(np.abs(coords1 - coords2)) > 3)

    def test_flipped_lactam(self):
        """
        Given a benzo + 7-membered lactam map a mol with the amide flipped
        """
        hit_F584 = TestSet.get_mol('F584')
        monster = Monster(hits=[hit_F584, ])
        monster.place_smiles(smiles='COc1cccc2C(=O)NCCCc12', long_name='flipped_F584')
        # monster.show_comparison()
        self.assertEqual(len(monster.get_mcs_mappings(followup=monster.initial_mol, hit=monster.hits[0])[0][0]), 13)

    def test_forced_flipped_lactam(self):

        """
        Given a benzo + 7-membered lactam map a mol with the amide flipped as the previous test
        force the ketones to match
        """
        hit_F584 = TestSet.get_mol('F584')
        hit_F584.SetProp('_Name', 'hit_F584')
        monster = Monster(hits=[hit_F584, ])
        monster.place_smiles(smiles='COc1cccc2C(=O)NCCCc12',
                             long_name='flipped_F584',
                             custom_map={'hit_F584': {13: 8}},
                             # merging_mode='expansion', # works on both expansion and no blend modes
                             )
        # monster.draw_nicely(monster.hits[0])
        # monster.draw_nicely(monster.initial_mol)
        # monster.show_comparison()
        # the [0][0][13] is because of:
        # Tuple[List[Dict[int, int]], ExtendedFMCSMode]
        self.assertIn(13, monster.get_mcs_mappings(followup=monster.initial_mol, hit=monster.hits[0])[0][0])
        self.assertEqual(monster.get_mcs_mappings(followup=monster.initial_mol, hit=monster.hits[0])[0][0][13], 8)
        # self.assertEqual(len(monster.get_mcs_mappings(followup=monster.initial_mol, hit=monster.hits[0])[0][0]), 13)

    def test_thiazole_flip(self):
        """
        See ``.get_5SB7_mols`` for details.

        Potentially the number of specified atoms is excessive,
        but as a human I just wrote down the atoms in the ring and was done with it.
        """
        mols = TestSet.get_5SB7_mols()
        monster = Monster([mols[0]])
        monster.place(Chem.Mol(mols[0]),
                      merging_mode='expansion',
                      # custom_map={'F36': {1:7}, 'F04': {4:7}}
                      custom_map={'F04': {-1: 4,  # no amine
                                          4: -2,  # no amine
                                          12: 12,
                                          6: 13,
                                          13: 6,
                                          15: 14,
                                          14: 15}}
                      )
        self.assertEqual(len(list(filter(len, monster.origin_from_mol(monster.positioned_mol)))), 15)
        # the amine is banned in the map and the ring is flipped.

    def test_thiazole_followup(self):
        """
        The followup compound is from 5SB3
        and is 'bold4' in mols.
        """
        mols = TestSet.get_5SB7_mols()
        monster = Monster(mols[:2])
        monster.place(Chem.Mol(mols[3]),
                      merging_mode='expansion',
                      custom_map={'F36': {1: 7},
                                  'F04': {4: -1,  # no amine
                                          12: 13,  # root to Ph
                                          13: 5,
                                          6: 14,
                                          }
                                  }
                      )
        # monster.show_comparison()
        # monster.to_nglview(True)
        # the amine at 7 is banned in F04.
        self.assertEqual(monster.origin_from_mol()[7], ['F36.1'])
        self.assertEqual(len(list(filter(len, monster.origin_from_mol(monster.positioned_mol)))), 22)

    def test_renumber(self):
        mols = TestSet.get_5SB7_mols()
        # F04 {root: 12, N: 13, amine: 0, C: 6},
        # original {root: 13, N: 5, amine: 7, C: 14},
        # renumbered {root: 8, N:19, amine: 12, C: 9}
        renumbered = Monster.renumber_followup_custom_map(mols[3],
                                                          Chem.MolFromSmiles(Chem.MolToSmiles(mols[3])),
                                                          custom_map={'F36': {1: 7},
                                                                      'F04': {0: -1,  # no amine
                                                                              -1: 7,
                                                                              13: 5,  # to N
                                                                              12: 13,  # root to Ph
                                                                              6: 14,  # to C
                                                                              }
                                                                      }

                                                          )
        expected = {'F36': {1: 12}, 'F04': {0: -1, -1: 12, 13: 19, 12: 8, 6: 9}}
        for name in renumbered:
            self.assertEqual(tuple(renumbered[name]), tuple(expected[name]),
                             f'{name} {renumbered[name]} != {expected[name]}')

    def test_victor_mol(self):
        """
        Victor from mol has different indices than from a smiles
        """
        mols = TestSet.get_5SB7_mols()
        Igor.init_pyrosetta()

        victor = Victor(hits=mols[:2], pdb_block=MPro.get_template(), ligand_resi='1X')

        victor.place(mols[3],
                     long_name=mols[3].GetProp('_Name'),
                     merging_mode='expansion',
                     custom_map={'F36': {1: 7},
                                 'F04': {0: -1,  # no amine
                                         -1: 7,  # no amine
                                         13: 5,  # to N
                                         12: 13,  # root to Ph
                                         6: 14,  # to C
                                         }
                                 }
                     )
        # victor.show_comparison()
        # victor.to_nglview()
        rsmd = victor.validate(mols[3])['reference2minimized_rmsd']
        self.assertLess(rsmd, 1.1, f"The resulting RMSD from the crystal is {rsmd}, which is greater than 1.")


if __name__ == '__main__':
    unittest.main()
